.parent_div {
	width: calc(1020px * 1.15);
	height: calc(700px * 1.15);
	position: absolute;
	margin: auto;
	top: 0;
	right: 0;
	bottom: 0;
	left: 0;
}

.child_div1 {
	width: calc(700px * 1.15);
	height: calc(700px * 1.15);
	float: left;
}

.canvas_container {
	width: calc(695px * 1.15);
	height: calc(695px * 1.15);
	border-style: dashed;
	border-width: 2px;
}

.child_div2 {
	width: calc(300px * 1.15);
	max-height: calc(700px * 1.15);
	float: right;
	overflow-y: scroll;
}

.controls_container {
	width: calc(300px * 1.15);
	position: relative;
}

.child_div2_child_mid {
	width: calc(300px * 1.15);
	height: 30px;
}

.infotext_container {
	width: calc(300px * 1.15);
}

p.sidebar-text {
	font-family: 'Open sans', sans-serif; color: #777777; font-size: 14px;
    hyphens: auto;
    text-align: justify;
}